/* 
* SDL (runtime and devel) is required to use this program.
* to compile under Linux, use the following command:
*
* gcc main.c -o points -lm `sdl-config --cflags --libs` -O3
* */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "SDL.h"


int place_points();
void add_to_list(Uint16 x, Uint16 y, Uint32 color);
void print_list();
Uint32 random_color(void);
void putpixel(SDL_Surface *surface, int x, int y, Uint32 pixel);
int fill_screen();
Uint32 find_nearest_color(int x, int y);
int dist_coords(int x1, int y1, int x2, int y2);

typedef struct _color_point {
		Uint16 x, y;
		Uint32 color;
		void *next;
} color_point;

SDL_Surface *screen;  // the global screen surface
color_point *cp_list=NULL;


int main(int argc, char **argv)
{
	int res;
		
	srand(time( NULL ));

	
	res = SDL_Init(SDL_INIT_VIDEO);
	if (res<0) { fprintf(stderr, "could not init video: %s\n", SDL_GetError()); return -1; }

	screen = SDL_SetVideoMode(512, 384, 24, SDL_SWSURFACE);
	if (screen == NULL) { fprintf(stderr, "could not set video mode: %s\n", SDL_GetError()); return
-1; }

	SDL_WM_SetCaption("points test by Raphael Assenat", NULL);

	place_points();	
	print_list();
	fill_screen();
		
	while (1)
	{
		SDL_Event ev;

		SDL_WaitEvent(&ev);
		if (ev.type==SDL_QUIT) { SDL_Quit(); return 0; }		
	}	
	SDL_Quit();	

	return 0;
}

int fill_screen()
{
	int x, y;

	printf("filling screen\n");
	

	for (y=0; y<screen->h; y++)
	{
		if ( SDL_MUSTLOCK(screen) ) {
			if (SDL_LockSurface(screen) < 0 ) {
				fprintf(stderr, "Cannot lock screen: %s\n", SDL_GetError());
				return 0;
			}
		}
		
		for (x=0; x<screen->w; x++)
		{					
			putpixel(screen, x, y, find_nearest_color(x, y));
		}
	
		if (SDL_MUSTLOCK(screen) ) {
			SDL_UnlockSurface(screen);
		}


		SDL_UpdateRect(screen, 0, y, screen->w, 1);
	}
	
	return 0;
}


int place_points()
{
	SDL_Event ev;
	while (1)
	{
		SDL_WaitEvent(&ev);

		switch (ev.type)
		{
			case SDL_QUIT: SDL_Quit(); exit(0); break;
			case SDL_MOUSEBUTTONDOWN:
						   				if (ev.button.button==SDL_BUTTON_LEFT)
										{
											add_to_list(ev.button.x, ev.button.y, random_color());
										}
										if (ev.button.button==SDL_BUTTON_RIGHT)
										{
											return 0;
										}
			
		}
	}
	return 0;
}

Uint32 random_color(void)
{
	Uint8 r, g, b;

	r = (int)(255.0*rand()/(RAND_MAX+1.0));
	g = (int)(255.0*rand()/(RAND_MAX+1.0));
	b = (int)(255.0*rand()/(RAND_MAX+1.0));

	
	return SDL_MapRGB(screen->format, r, g, b);
}

void add_to_list(Uint16 x, Uint16 y, Uint32 color)
{
	color_point *new_point=NULL;
	color_point *item;

	new_point = (color_point*)malloc(sizeof(color_point));
	new_point->next = NULL;
	new_point->x = x;
	new_point->y = y;
	new_point->color = color;
		
	if (cp_list==NULL) { 
		cp_list = new_point; 
	} else {		
		item = cp_list;
		while (item->next!=NULL) { item = item->next; }
		item->next = new_point;
	}

	if ( SDL_MUSTLOCK(screen) ) {
		if (SDL_LockSurface(screen) < 0 ) {
			fprintf(stderr, "Cannot lock screen: %s\n", SDL_GetError());
			return;
		}
	}

	putpixel(screen, x, y, color);

	if ((x<screen->w)&&(y<screen->h)) {	
		putpixel(screen, x+1, y, color);
		putpixel(screen, x+1, y+1, color);
		putpixel(screen, x, y+1, color);
	}
	
	if (SDL_MUSTLOCK(screen) ) {
		SDL_UnlockSurface(screen);
	}

	SDL_UpdateRect(screen, x, y, 2, 2);
//	printf("added %d, %d\n", x, y);
}

void print_list()
{
	color_point *item;
	item = cp_list;

	if (item==NULL) { return; }

	while (item->next!=NULL) {
		printf("X: %d   Y: %d   Color: %d\n", item->x, item->y, item->color);
		item = item->next;
	}
	printf("X: %d   Y: %d   Color: %d\n", item->x, item->y, item->color);
}

Uint32 find_nearest_color(int x, int y)
{
	color_point *item = cp_list;
	color_point *best_found = cp_list;
	int lower_dist;

	lower_dist = dist_coords(x, y, cp_list->x, cp_list->y);
	
	if (item==NULL) { return 0; }	

	while (item->next!=NULL) {
		if (dist_coords(x, y, item->x, item->y)<lower_dist)
		{
			lower_dist = dist_coords(x, y, item->x, item->y);
			best_found = item;
		}
		item = item->next;
	}

	if (dist_coords(x, y, item->x, item->y)<lower_dist)
	{
		lower_dist = dist_coords(x, y, item->x, item->y);
		best_found = item;
	}
	
	return best_found->color;	
}

int dist_coords(int x1, int y1, int x2, int y2)
{
	int result=(int)hypot( abs(x1-x2),abs(y1-y2) );
	//return sin(log(pow(result, 3)))*(double)result;
	//return (int)(     log(abs((x1-x2))+abs(y1-y2))*20.0    );
	return result;
}

void putpixel(SDL_Surface *surface, int x, int y, Uint32 pixel)
{
	int bpp = surface->format->BytesPerPixel;

	Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

	switch(bpp) {
		case 1:
			* p = pixel;
			break;

		case 2:
			*(Uint16*)p = pixel;
			break;

		case 3:
			if (SDL_BYTEORDER == SDL_BIG_ENDIAN) {
				p[0] = (pixel >> 16) & 0xff;
				p[1] = (pixel >> 8) & 0xff;
				p[2] = pixel & 0xff;
			} else {
				p[0] = pixel & 0xff;
				p[1] = (pixel >> 8) & 0xff;
				p[2] = (pixel >>16) & 0xff;
			}
			break;
		
		case 4:
			*(Uint32*)p = pixel;
			break;
	}
}

